var init = (function (glib2, gvc1) {
    'use strict';

    let timeoutSourceIds = [];
    function delay(milliseconds) {
        return new Promise((resolve) => {
            const timeoutId = glib2.timeout_add(glib2.PRIORITY_DEFAULT, milliseconds, () => {
                removeFinishedTimeoutId(timeoutId);
                resolve(undefined);
                return glib2.SOURCE_REMOVE;
            });
            if (!timeoutSourceIds) {
                timeoutSourceIds = [];
            }
            timeoutSourceIds.push(timeoutId);
        });
    }
    function removeFinishedTimeoutId(timeoutId) {
        timeoutSourceIds?.splice(timeoutSourceIds.indexOf(timeoutId), 1);
    }
    function disposeDelayTimeouts() {
        timeoutSourceIds?.forEach((sourceId) => {
            glib2.Source.remove(sourceId);
        });
        timeoutSourceIds = null;
    }

    const Main = imports.ui.main;
    const QuickSettings = Main.panel.statusArea.quickSettings;
    class AudioPanel {
        getDisplayedDevices() {
            const devices = QuickSettings._volume._output._deviceItems;
            return Array.from(devices, ([id, value]) => ({
                id,
                displayName: value.label.get_text(),
            }));
        }
        removeDevice(id) {
            QuickSettings._volume._output._removeDevice(id);
        }
        addDevice(id) {
            QuickSettings._volume._output._addDevice(id);
        }
    }

    function range(amount) {
        return [...Array(amount).keys()];
    }

    /**
     * Display name format copied from
     * https://gitlab.gnome.org/GNOME/gnome-shell/-/blob/main/js/ui/status/volume.js#L132
     * The "-" is U+2013 on purpose
     * @param device AudioDevice
     */
    function getAudioDevice(id, description, origin) {
        if (!description)
            description = "unknown description";
        if (!origin)
            origin = "unknown origin";
        return {
            id,
            displayName: `${description} – ${origin}`,
        };
    }

    class MixerWrapper {
        constructor(mixer, disposal) {
            this.mixer = mixer;
            this.disposal = disposal;
        }
        getAudioDevicesFromIds(ids) {
            return ids.map((id) => {
                const lookup = this.mixer.lookup_output_id(id);
                return getAudioDevice(id, lookup?.get_description(), lookup?.get_origin());
            });
        }
        /**
         * Uses a Dummy Device "trick" from
         * https://github.com/kgshank/gse-sound-output-device-chooser/blob/master/sound-output-device-chooser@kgshank.net/base.js#LL299C20-L299C20
         * @param displayNames display names
         * @returns A list of matching audio devices. If a given display name is not found,
         * undefined is returned in its place.
         */
        getAudioDevicesFromDisplayNames(displayNames) {
            const dummyDevice = new gvc1.MixerUIDevice();
            const devices = this.getAudioDevicesFromIds(range(dummyDevice.get_id()));
            return displayNames.map((name) => devices.find((device) => device.displayName === name));
        }
        subscribeToOutputChanges(callback) {
            const addId = this.mixer.connect("output-added", (_, deviceId) => callback({ deviceId, type: "output-added" }));
            const removeId = this.mixer.connect("output-removed", (_, deviceId) => callback({ deviceId, type: "output-removed" }));
            return { ids: [addId, removeId] };
        }
        unsubscribe(subscription) {
            subscription.ids.forEach((id) => {
                this.mixer.disconnect(id);
            });
        }
        dispose() {
            this.disposal();
        }
    }

    async function waitForMixerToBeReady(mixer) {
        while (mixer.get_state() === gvc1.MixerControlState.CONNECTING) {
            await delay(200);
        }
        const state = mixer.get_state();
        if (state === gvc1.MixerControlState.FAILED) {
            throw new Error("MixerControl is in a failed state");
        }
        else if (state === gvc1.MixerControlState.CLOSED) {
            throw new Error("MixerControl is in a closed state");
        }
    }

    class AudioPanelMixerSource {
        async getMixer() {
            const Volume = imports.ui.status.volume;
            const mixer = Volume.getMixerControl();
            await waitForMixerToBeReady(mixer);
            await delay(200);
            return new MixerWrapper(mixer, () => { });
        }
    }

    const ExtensionUtils = imports.misc.extensionUtils;
    const SettingsPath = "org.gnome.shell.extensions.quicksettings-audio-devices-hider";
    const ExcludedOutputNamesSetting = "excluded-output-names";
    const AvailableOutputNames = "available-output-names";
    class SettingsUtils {
        constructor() {
            this.settings = null;
        }
        getSettings() {
            if (!this.settings) {
                this.settings = ExtensionUtils.getSettings(SettingsPath);
            }
            return this.settings;
        }
        getExcludedOutputDeviceNames() {
            const settings = this.getSettings();
            const ids = settings.get_strv(ExcludedOutputNamesSetting);
            return ids;
        }
        setExcludedOutputDeviceNames(displayNames) {
            const settings = this.getSettings();
            settings.set_strv(ExcludedOutputNamesSetting, displayNames);
        }
        addToExcludedOutputDeviceNames(displayName) {
            const currentOutputs = this.getExcludedOutputDeviceNames();
            if (currentOutputs.includes(displayName)) {
                return;
            }
            const newOutputs = [...currentOutputs, displayName];
            const settings = this.getSettings();
            settings.set_strv(ExcludedOutputNamesSetting, newOutputs);
        }
        removeFromExcludedOutputDeviceNames(displayName) {
            const outputs = this.getExcludedOutputDeviceNames();
            const index = outputs.indexOf(displayName);
            if (index === -1) {
                return;
            }
            outputs.splice(index, 1);
            const settings = this.getSettings();
            settings.set_strv(ExcludedOutputNamesSetting, outputs);
        }
        getAvailableOutputs() {
            const settings = this.getSettings();
            const ids = settings.get_strv(AvailableOutputNames);
            return ids;
        }
        setAvailableOutputs(displayNames) {
            const settings = this.getSettings();
            settings.set_strv(AvailableOutputNames, displayNames.map((id) => id.toString()));
        }
        addToAvailableOutputs(displayName) {
            const currentOutputs = this.getAvailableOutputs();
            if (currentOutputs.includes(displayName)) {
                return;
            }
            const newAllOutputs = [...currentOutputs, displayName];
            const settings = this.getSettings();
            settings.set_strv(AvailableOutputNames, newAllOutputs.map((id) => id.toString()));
        }
        removeFromAvailableOutputs(displayName) {
            const outputs = this.getAvailableOutputs();
            const index = outputs.indexOf(displayName);
            if (index === -1) {
                return;
            }
            outputs.splice(index, 1);
            const settings = this.getSettings();
            settings.set_strv(AvailableOutputNames, outputs.map((id) => id.toString()));
        }
        connectToChanges(settingName, func) {
            return this.getSettings().connect(`changed::${settingName}`, func);
        }
        disconnect(subscriptionId) {
            this.getSettings().disconnect(subscriptionId);
        }
        dispose() {
            this.settings = null;
        }
    }

    imports.misc.extensionUtils;
    class Extension {
        constructor(uuid) {
            this._uuid = uuid;
        }
        enable() {
            log(`Enabling extension ${this._uuid}`);
            this._audioPanel = new AudioPanel();
            this._settings = new SettingsUtils();
            this._lastExludedDevices = this._settings.getExcludedOutputDeviceNames();
            new AudioPanelMixerSource().getMixer().then((mixer) => {
                this._mixer = mixer;
                this.setAllOutputsInSettings();
                this.setupOutputChangesSubscription();
                this.hideExcludedDevices();
                this.setupExludedDevicesHandling();
            });
        }
        hideExcludedDevices() {
            var devices = this._mixer?.getAudioDevicesFromDisplayNames(this._lastExludedDevices);
            devices?.forEach((device) => {
                if (device) {
                    this._audioPanel.removeDevice(device.id);
                }
            });
        }
        setupExludedDevicesHandling() {
            this._settingsSubscription = this._settings.connectToChanges(ExcludedOutputNamesSetting, () => {
                const newExcludedDevices = this._settings?.getExcludedOutputDeviceNames();
                if (!newExcludedDevices) {
                    return;
                }
                const devicesToShowIds = this.getDeviceIdsToShow(newExcludedDevices);
                const devicesToHideIds = this.getDeviceIdsToHide(newExcludedDevices);
                devicesToShowIds?.forEach((id) => this._audioPanel?.addDevice(id));
                devicesToHideIds?.forEach((id) => this._audioPanel?.removeDevice(id));
                this._lastExludedDevices = newExcludedDevices ?? [];
            });
        }
        getDeviceIdsToHide(newExcludedDevices) {
            const devicesToHide = newExcludedDevices.filter((current) => !this._lastExludedDevices.includes(current));
            const devicesToHideIds = this._mixer
                ?.getAudioDevicesFromDisplayNames(devicesToHide)
                .filter((n) => n)
                .map((n) => n.id);
            return devicesToHideIds;
        }
        getDeviceIdsToShow(newExcludedDevices) {
            const devicesToShow = this._lastExludedDevices.filter((last) => !newExcludedDevices.includes(last));
            const devicesToShowIds = this._mixer
                ?.getAudioDevicesFromDisplayNames(devicesToShow)
                .filter((n) => n)
                .map((n) => n.id);
            return devicesToShowIds;
        }
        setupOutputChangesSubscription() {
            this._mixerSubscription =
                this._mixer?.subscribeToOutputChanges((event) => {
                    this.updateAvailableOutputsInSettings(event);
                    if (event.type === "output-added") {
                        this.hideDeviceIfExcluded(event.deviceId);
                    }
                }) ?? null;
        }
        hideDeviceIfExcluded(deviceId) {
            if (!this._mixer) {
                return;
            }
            const deviceName = this._mixer.getAudioDevicesFromIds([deviceId])[0]
                .displayName;
            const excludedOutputs = this._settings?.getExcludedOutputDeviceNames();
            if (excludedOutputs?.includes(deviceName)) {
                delay(200).then(() => {
                    // delay due to potential race condition with Quick Setting panel's code
                    this._audioPanel.removeDevice(deviceId);
                });
            }
        }
        setAllOutputsInSettings() {
            const allDisplayedDevices = this._audioPanel.getDisplayedDevices();
            this._settings.setAvailableOutputs(allDisplayedDevices.map((i) => i.displayName));
        }
        updateAvailableOutputsInSettings(event) {
            if (!this._mixer) {
                return;
            }
            const displayName = this._mixer.getAudioDevicesFromIds([event.deviceId])[0]
                .displayName;
            if (event.type === "output-added") {
                this._settings.addToAvailableOutputs(displayName);
            }
            else if (event.type === "output-removed") {
                this._settings.removeFromAvailableOutputs(displayName);
            }
            else {
                log(`WARN: Received an unsupported MixerEvent: ${event.type}`);
            }
        }
        disable() {
            log(`Disabling extension ${this._uuid}`);
            if (this._mixerSubscription) {
                this._mixer?.unsubscribe(this._mixerSubscription);
            }
            this._mixer?.dispose();
            if (this._settingsSubscription) {
                this._settings?.disconnect(this._settingsSubscription);
                this._settingsSubscription = null;
            }
            this.enableAllDevices();
            disposeDelayTimeouts();
            this._settings?.dispose();
            this._settings = null;
            this._audioPanel = null;
            this._lastExludedDevices = null;
            this._mixer = null;
            this._mixerSubscription = null;
        }
        enableAllDevices() {
            const allDevices = this._settings.getAvailableOutputs();
            const devicesToShowIds = this._mixer
                ?.getAudioDevicesFromDisplayNames(allDevices)
                .filter((n) => n)
                .map((n) => n.id);
            devicesToShowIds?.forEach((id) => this._audioPanel.addDevice(id));
        }
    }
    function extension (meta) {
        return new Extension(meta.uuid);
    }

    return extension;

})(imports.gi.GLib, imports.gi.Gvc);
